/*
 *  Copyright (C) 2010  Regents of the University of Michigan
 *
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _PILEUP_READER_H
#define _PILEUP_READER_H

#include <stdint.h>
#include <string>

#include "SamFile.h"

///
/// \file this file implements a means of reading the
/// bases and structural variations of interest that are
/// present in any BAM file.
/// 
/// Two data structures contain the pertinent information:
///
/// - ReadBase contains information about a single base in
///   a single sequence, and is a copy of the useful fields in
///   the BAM record that it originated from.
///
/// - ReadInsertion contains the information that can't
///   be represented as a base itself, which is material
///   inserted at the genome at a given location.
///
/// The class ReadBase represents either bases that are directly
/// present in the class or it represents deletions of those
/// bases.  So for example, if you ask for chromosome 1, position
/// 25 and the corresponding aligned read indicates in the CIGAR
/// string that there is a deletion there, the ReadBase object
/// will not show a valid base, but rather indicate that this
/// base and some number of subsequent bases were deleted.
///
/// Insertions cannot be represented directly as a single added
/// base nor as a set of bases that were deleted (which ReadBase
/// does) but rather a set of bases that appears in between two
/// positions on the reference.
///
/// The PileupReader class provides the mechanism for attaching
/// to a BAM file and retrieving the above data.
///
/// Use cases:
///
/// - read the BAM data associated with specific known marker
///   positions, usually those read locations that are provided
///   by a sequencing chip.  In this case, PileupReader will by
///   default use an index if present to rapidly skip from marker
///   to marker.  This behavior may be controlled by using the
///   useIndex() method.
///
/// - scan the entire genome, base by base, reading bases and
///   insertions and evaluating them, for example to do novel
///   SNP detection.
/// 

///
/// When we call PileupReader::getPileup, we return a set
/// of these objects representing the bases found in all
/// overlapping reads in the given BAM file.
///
/// The information in this class is the pertinent information
/// from the SAM format record.
///
class ReadBase {
public:
    ReadBase() :
        base('N'),
        phredQuality('!'),
        mapQuality(0),
        SAMFlag(0),
        deletedBaseCount(0),
        indelInRead(false)
        {;}
    char    base;               // 'ACGT', '0123", 'N' only
    char    phredQuality;       // ASCII character '!' to '~' inclusive.
    uint8_t mapQuality;
    uint16_t SAMFlag;
    int     deletedBaseCount;   // how many bases, including this one, are deleted
    int     insertedBasesOverlapping;
    bool    indelInRead;        // the read contains an indel
};

///
/// Contain all pertinent information from overlapping
/// SAM records that START an insertion at this point.
///
class ReadInsertion {
public:
    ReadInsertion() :
        mapQuality(0),
        forwardStrand(true) {;}
    std::string bases;
    std::string phredQualities;
    uint8_t mapQuality;
    bool    forwardStrand;
};

///
/// class for tracking pileups in a BAM file.
///
/// Given a sorted BAM file, allow us to scan over the file, either base by
/// base or marker by marker in ascending order, and obtain a list of
/// bases from the BAM file at the corresponding position.
///
/// In the even that an insertion is noticed after the marker of interest,
/// signal that in getPileup, and allow the data to be returned by calling
/// getInsertions().
///
///
class PileupReader {

private:
    SamFile&    _samFile;
    int         _chromosome;
    uint32_t    _position;
    bool        _useIndex;
    std::vector<SamRecord *>    _samRecords;    // vector of ptrs to recovers that overlap the last pileup position

    bool    _getPileupOverflow;
    bool    _getInsertionsOverflow;

public:

    PileupReader(SamFile &samFile) :
        _samFile(samFile),
        _chromosome(0),
        _position(0),
        _useIndex(true),
        _getPileupOverflow(false),
        _getInsertionsOverflow(false)
        { ; }

    /// \brief indicate whether the method getPileup() holds
    /// fewer bases than were actually present in the sequencing
    /// file.
    bool didGetPileupOverflow() { return _getPileupOverflow; }

    /// \brief indicate whether the method getInsertions() holds
    /// fewer inserted sequences than were actually present in
    /// the sequencing file.
    bool didGetInsertionsOverflow() { return _getInsertionsOverflow; }

    /// \brief clear the pileup overflow flag
    void clearPileupOverflow() { _getPileupOverflow = false; }

    /// \brief clear the insertion overflow flag
    void clearInsertionsOverflow() { _getInsertionsOverflow = false; }

    void useIndex(bool u=true) { _useIndex = u; }

    /// \brief getPileup returns a pileup vector of the same
    /// length as the number of reads that overlaps the given
    /// chromosome and position
    /// 
    /// \param chromosome  identifies the chromosome in which
    /// to retrieve bases from.
    /// 
    /// \param position  indicates the 1-based offset of the 
    /// base of interest on that chromosome.  This number is
    /// relative to the genome reference that the BAM file
    /// was aligned against.
    /// 
    /// \param readBases a caller allocated array of of readBase
    /// objects, of sufficient size to contain read depths of interest.
    /// 
    /// \param baseCount on input, this is the maxumum allocated size
    /// of the memory pointed to by readBases.  On exit, it contains the
    /// number of bases actually read.
    /// 
    /// \param basesInserted bool that indicates if getInsertions() will
    /// return a non-zero vector of bases that were inserted AFTER
    /// this position.  Since reads may truncate this insertion,
    /// they may be different lengths.
    ///
    /// \return true if more data exists in the BAM file, false otherwise
    ///
    bool getPileup(int chromosome, uint32_t position, ReadBase *readBases, int &baseCount, bool &basesInserted);

    bool getInsertions(int chromosome, uint32_t position, ReadInsertion *readInsertions, int &insertionsCount);

};


#endif
